"""Basic functional tests for all of the kvlayer backends.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

"""
from __future__ import absolute_import
import logging
import random
from StringIO import StringIO  # cStringIO is not pickable
import uuid

import pkg_resources
import pytest
import yaml

from pytest_diffeo import redis_address
import kvlayer
from kvlayer import BadKey
from kvlayer._client import STORAGE_CLIENTS, load_entry_point_kvlayer_impls
import yakonfig

logger = logging.getLogger(__name__)


_extension_test_configs = {}


def load_entry_point_kvlayer_test_configs():
    global _extension_test_configs
    for entry_point in pkg_resources.iter_entry_points('kvlayer.test_config'):
        try:
            name = entry_point.name
            constructor = entry_point.load()
            _extension_test_configs[name] = constructor()
        except:
            logger.error('failed loading kvlayer test_config %r',
                         entry_point and entry_point.name, exc_info=True)


# run at global scope to ensure it's before we might possibly need
# STORAGE_CLIENTS in any context
load_entry_point_kvlayer_impls()
load_entry_point_kvlayer_test_configs()


@pytest.fixture(scope='module',
                params=STORAGE_CLIENTS.keys())
def backend(request):
    backend = request.param
    if backend == 'cassandra':
        pytest.skip('cassandra doesn\'t support non-UUID keys')
    if backend in _extension_test_configs:
        pass  # okay
    elif not request.fspath.dirpath('config_{}.yaml'.format(backend)).exists():
        pytest.skip('no configuration file for backend {}'.format(backend))
    #if backend not in ('local', 'redis', 'riak', 'accumulo', 'cborproxy'):
    #if backend != 'local':
    #    pytest.skip('TODO DELETE TEMPROARY SKIP ' + backend)
    return backend


@pytest.yield_fixture(scope='function')
def client(backend, request, tmpdir, namespace_string):
    if backend in _extension_test_configs:
        file_config = yaml.load(_extension_test_configs[backend])
    else:
        config_path = str(request.fspath.dirpath('config_{}.yaml'
                                                 .format(backend)))
        # read and parse the config file, insert an object
        with open(config_path, 'r') as f:
            file_config = yaml.load(f)

    # Insert an object into the config which stats will write to.
    # Below we can get the stats text and log it here.
    # (Normal stats flow logs to file.)
    file_config['kvlayer']['log_stats'] = StringIO()
    file_config['kvlayer']['encoder'] = 'packed'

    params = dict(
        app_name='kvlayer',
        namespace=namespace_string,
    )

    # this is hacky but must go somewhere
    if backend == 'filestorage':
        local = tmpdir.join('local')
        with local.open('w') as f:
            pass
        params['kvlayer_filename'] = str(local)

    if backend == 'redis':
        params['storage_addresses'] = [redis_address(request)]

    with yakonfig.defaulted_config(
            [kvlayer],
            config=file_config,
            params=params):
        client = kvlayer.client()
        client.delete_namespace()
        yield client
        if client._log_stats is not None:
            client._log_stats.flush()
            logger.info('storage stats (%s %s):\n%s',
                        backend, request.function.__name__,
                        file_config['kvlayer']['log_stats'].getvalue())
        client.delete_namespace()


def test_basic_storage(client):
    client.setup_namespace(dict(t1=2, t2=3))
    assert 0 == len(list(client.scan('t1')))
    # use time-based UUID 1, so these are ordered
    u1, u2, u3 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1()
    client.put('t1', ((u1, u2), b'88'))
    client.put('t2', ((u1, u2, u3), b'88'))
    assert 1 == len(list(client.scan('t1')))
    assert 1 == len(list(client.scan('t1', ((u1,), (u1,)))))

    client.delete('t1', (u1, u2))
    assert 0 == len(list(client.scan('t1')))
    assert 0 == len(list(client.scan('t1', ((u1,), (u1,)))))
    assert 0 == len(list(client.scan('t2', ((u2,), (u3,)))))


def test_delete(client):
    client.setup_namespace({'table1': 1})
    kv_dict = {(uuid.uuid4(),): 'value' + str(x) for x in xrange(10)}
    keys_and_values = [(key, value) for key, value in kv_dict.iteritems()]
    client.put('table1', *keys_and_values)
    keys = kv_dict.keys()
    delete_keys = keys[0:len(keys):2]
    save_keys = keys[1:len(keys):2]
    assert set(delete_keys)
    client.delete('table1', *delete_keys)
    for key in save_keys:
        for key, value in client.get('table1', key):
            assert kv_dict[key] == value
    for key in delete_keys:
        assert list(client.get('table1', key)) == [(key, None)]


def test_get(client):
    client.setup_namespace(dict(t1=1, t2=2, t3=3))
    # use time-based UUID 1, so these are ordered
    u1, u2, u3 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1()

    client.put('t1', ((u1,), b'81'))
    res = list(client.get('t1', (u1, )))
    assert len(res) == 1
    assert res[0][0] == (u1,)
    assert res[0][1] == b'81'

    client.put('t2', ((u1, u2), b'82'))
    res = list(client.get('t2', (u1, u2)))
    assert len(res) == 1
    assert res[0][0] == (u1, u2)
    assert res[0][1] == b'82'

    client.put('t3', ((u1, u2, u3), b'83'))
    res = list(client.get('t3', (u1, u2, u3)))
    assert len(res) == 1
    assert res[0][0] == (u1, u2, u3)
    assert res[0][1] == b'83'


def test_put_put(client):
    client.setup_namespace(dict(t1=1))
    u1 = uuid.uuid1()
    client.put('t1', ((u1,), 'a'))
    client.put('t1', ((u1,), 'b'))
    assert list(client.get('t1', (u1,))) == [((u1,), 'b')]


def test_adding_tables(client):
    client.setup_namespace(dict(t1=2, t2=3))
    # use time-based UUID 1, so these are ordered
    u1, u2, u3 = uuid.uuid1(), uuid.uuid1(), uuid.uuid1()
    client.put('t1', ((u1, u2), b'11'))
    client.put('t2', ((u1, u2, u3), b'22'))

    client.setup_namespace(dict(t3=1))
    client.put('t3', ((u1,), b'33'))

    assert 1 == len(list(client.scan('t1')))
    assert 1 == len(list(client.scan('t1', ((u1,), (u1,)))))
    assert 1 == len(list(client.scan('t3', ((u1,), (u1,)))))
    assert 0 == len(list(client.scan('t2', ((u2,), (u3,)))))


def test_setup_namespace_idempotent(client):
    client.setup_namespace(dict(t1=2))
    u1, u2 = uuid.uuid1(), uuid.uuid1()
    client.put('t1', ((u1, u2), b'88'))
    assert 1 == len(list(client.scan('t1')))
    assert 1 == len(list(client.scan('t1', ((u1,), (u1,)))))

    client.setup_namespace(dict(t1=2))
    assert 1 == len(list(client.scan('t1')))
    assert 1 == len(list(client.scan('t1', ((u1,), (u1,)))))

    client.delete_namespace()
    client.setup_namespace(dict(t1=2))
    assert 0 == len(list(client.scan('t1')))
    assert 0 == len(list(client.scan('t1', ((u1,), (u1,)))))


def test_clear_table(client):
    client.setup_namespace(dict(t1=2, t2=3))
    num_rows = 10 ** 2
    # make two tables and reset only one
    client.put('t1', *[((uuid.uuid4(), uuid.uuid4()), b'')
                       for i in xrange(num_rows)])
    client.put('t2', *[((uuid.uuid4(), uuid.uuid4(),
                         uuid.uuid4()), b'')
                       for i in xrange(num_rows)])
    client.clear_table('t1')
    assert len(list(client.scan('t1'))) == 0
    assert len(list(client.scan('t2'))) == num_rows

    client.put('t1', *[((uuid.uuid4(), uuid.uuid4()), b'')
                       for i in xrange(num_rows)])
    assert len(list(client.scan('t1'))) == num_rows


def test_bogus_put(client):
    client.setup_namespace(dict(t1=2, t2=3))
    num_rows = 10 ** 2
    # make two tables and reset only one
    with pytest.raises(BadKey):
        client.put('t1', *[((uuid.uuid4(), uuid.uuid4(),
                             uuid.uuid4(), uuid.uuid4()), b'')
                           for i in xrange(num_rows)])
    with pytest.raises(BadKey):
        client.put('t2', *[((uuid.uuid4(), uuid.uuid4(),
                             uuid.uuid4(), uuid.uuid4()), b'')
                           for i in xrange(num_rows)])


def sequence_to_one(iterable):
    val = None
    for tv in iterable:
        assert val is None
        val = tv
    assert val is not None
    return val


def test_binary_clean(client):
    '''Test binary integrity moving non-text bytes to db and back.'''
    client.setup_namespace(dict(t1=2))
    keya = (uuid.uuid4(), uuid.uuid4())
    # every byte value from 0..255
    vala = bytes(b''.join([chr(x) for x in xrange(0, 256)]))
    client.put('t1', (keya, vala))
    keyb = (uuid.uuid4(), uuid.uuid4())
    valb = bytes(b''.join([chr(random.randint(0, 255)) for x in xrange(1000)]))
    client.put('t1', (keyb, valb))
    xvala = sequence_to_one(client.scan('t1', (keya, keya)))
    assert xvala[1] == vala
    xvalb = sequence_to_one(client.scan('t1', (keyb, keyb)))
    assert xvalb[1] == valb


def test_scan(client):
    client.setup_namespace(dict(t1=2))

    # Add all keys and values from 0 to 90 counting by 10s.
    for x in xrange(0, 100, 10):
        key = (uuid.UUID(int=x), uuid.UUID(int=x))
        client.put('t1', (key, '%d' % x))

    # Scan entire range
    expected_results = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90]
    actual_results = [int(v) for k, v in client.scan('t1', ((), ()))]
    assert actual_results == expected_results

    # Scan with specified start and stop values
    expected_results = [40, 50, 60, 70]
    actual_results = [int(v) for k, v in
                      client.scan('t1', ((uuid.UUID(int=40),),
                                         (uuid.UUID(int=75),)))]
    assert actual_results == expected_results

    # Scan with specified start and stop values
    # 40 <= x <= 70
    expected_results = [40, 50, 60, 70]
    actual_results = [int(v) for k, v in
                      client.scan('t1', ((uuid.UUID(int=40),),
                                         (uuid.UUID(int=70),)))]
    assert actual_results == expected_results

    # Scan to unspecified end of range from start key
    expected_results = [80, 90]
    actual_results = [int(v) for k, v in
                      client.scan('t1', ((uuid.UUID(int=80),), ()))]
    assert actual_results == expected_results

    # Scan from minimum value to specified middle value
    expected_results = [0, 10]
    actual_results = [int(v) for k, v in
                      client.scan('t1', ((uuid.UUID(int=0),),
                                         (uuid.UUID(int=15),)))]
    assert actual_results == expected_results

    # Scan from unspecified start value to middle value
    expected_results = [0, 10, 20, 30, 40, 50, 60, 70]
    actual_results = [int(v) for k, v in
                      client.scan('t1', ((), (uuid.UUID(int=75),)))]
    assert actual_results == expected_results


def test_scan_2d(client):
    '''Set up a table with keys x,y and scan a single known x.'''
    client.setup_namespace({'t1': 2})

    def kf(x, y):
        return (uuid.UUID(int=x), uuid.UUID(int=y))

    def k1f(x):
        return (uuid.UUID(int=x),)

    def vf(x, y):
        return '{}:{}'.format(x, y)

    def kvf(x, y):
        return (kf(x, y), vf(x, y))

    def scan(fr, to):
        return list(client.scan('t1', (fr, to)))

    r = range(1, 5)  # [1, 2, 3, 4]

    kvps = [kvf(x, y) for x in r for y in r]
    client.put('t1', *kvps)

    assert list(client.scan('t1')) == kvps
    assert scan(k1f(2), k1f(2)) == [kvf(2, y) for y in r]
    assert scan(kf(2, 1), kf(2, 2)) == [kvf(2, 1), kvf(2, 2)]
    assert scan(kf(2, 0), kf(2, 3)) == [kvf(2, 1), kvf(2, 2), kvf(2, 3)]
    assert scan(kf(2, 2), kf(2, 5)) == [kvf(2, 2), kvf(2, 3), kvf(2, 4)]
    assert scan(kf(2, 5), kf(3, 0)) == []
    assert scan(kf(2, 4), kf(3, 1)) == [kvf(2, 4), kvf(3, 1)]
    assert scan(k1f(4), k1f(6)) == [kvf(4, y) for y in r]
    assert scan(k1f(0), k1f(1)) == [kvf(1, y) for y in r]
    assert scan(k1f(0), k1f(6)) == kvps
    assert scan(kf(0, 6), kf(6, 0)) == kvps
    assert scan(k1f(0), k1f(0)) == []


def test_scan_keys_2d(client):
    '''Same as test_scan_2d, but only scan the key space.'''
    client.setup_namespace({'t1': 2})

    def kf(x, y):
        return (uuid.UUID(int=x), uuid.UUID(int=y))

    def k1f(x):
        return (uuid.UUID(int=x),)

    def vf(x, y):
        return '{}:{}'.format(x, y)

    def kvf(x, y):
        return (kf(x, y), vf(x, y))

    def scan_keys(fr, to):
        return list(client.scan_keys('t1', (fr, to)))

    r = range(1, 5)  # [1, 2, 3, 4]

    kvps = [kvf(x, y) for x in r for y in r]
    keys = [kf(x, y) for x in r for y in r]
    client.put('t1', *kvps)

    assert list(client.scan_keys('t1')) == keys
    assert scan_keys(k1f(2), k1f(2)) == [kf(2, y) for y in r]
    assert scan_keys(kf(2, 1), kf(2, 2)) == [kf(2, 1), kf(2, 2)]
    assert scan_keys(kf(2, 0), kf(2, 3)) == [kf(2, 1), kf(2, 2), kf(2, 3)]
    assert scan_keys(kf(2, 2), kf(2, 5)) == [kf(2, 2), kf(2, 3), kf(2, 4)]
    assert scan_keys(kf(2, 5), kf(3, 0)) == []
    assert scan_keys(kf(2, 4), kf(3, 1)) == [kf(2, 4), kf(3, 1)]
    assert scan_keys(k1f(4), k1f(6)) == [kf(4, y) for y in r]
    assert scan_keys(k1f(0), k1f(1)) == [kf(1, y) for y in r]
    assert scan_keys(k1f(0), k1f(6)) == keys
    assert scan_keys(kf(0, 6), kf(6, 0)) == keys
    assert scan_keys(k1f(0), k1f(0)) == []


def test_scan_2d_prefix(client):
    '''scan(('a',),('a',)) shouldn't find ('ab','c')'''
    client.setup_namespace({'ts2': (str, str)})
    client.put('ts2',
               (('a', 'a'), '1'),
               (('a', 'z'), '2'),
               (('ab', 'b'), '3'),
               (('ab', 'y'), '4'))

    assert (list(client.scan('ts2', (('a',), ('a',)))) ==
            [(('a', 'a'), '1'), (('a', 'z'), '2')])  # but not anything ab
    assert (list(client.scan_keys('ts2', (('a',), ('a',)))) ==
            [('a', 'a'), ('a', 'z')])  # but not anything ab


def test_scan_9042(client):
    tname = 'ts9000'
    client.setup_namespace({tname: (str,)})

    kv = [
        (('k{:5x}'.format(x),), 'v{:5x}'.format(x))
        for x in xrange(9042)
    ]
    client.put(tname, *kv)

    # now check that scan gets them all back
    count = 0
    for rkv in client.scan(tname):
        count += 1

    assert count == 9042

    # try other ranged scans
    count = 0
    for rkv in client.scan(tname, (('k{:5x}'.format(3000),), ())):
        count += 1

    assert count == 6042

    count = 0
    for rkv in client.scan(tname, (('k{:5x}'.format(3000),),
                                   ('k{:5x}'.format(8437),))):
        count += 1

    assert count == 8437-3000+1  # +1 because scan is inclusive of endpoints

    count = 0
    for rkv in client.scan(tname, ((), ('k{:5x}'.format(8437),))):
        count += 1

    assert count == 8437+1  # +1 because scan is inclusive of endpoints


def test_scan_binary_key_order(client):
    client.setup_namespace({'s1': (str, int)})
    # keys in what should be sorted order.
    keys = [
        ('\0', 1),
        ('\0\0', 2),
        ('\0\x01', 3),
        ('\x01', 4),
        ('\x01\0', 5),
        ('\x01#aoeu', 6),
        ('\x02', 7),
    ]
    values = ['{:05d}'.format(i) for i in xrange(1, len(keys)+1)]
    client.put(
        's1',
        *zip(keys, values)
    )

    d = list(client.scan('s1'))
    dvalues = [dx[1] for dx in d]
    # ensure that values come back out in expected order.
    assert dvalues == values

    d = list(client.scan('s1', (('\x01',), None)))
    assert len(d) == 4


def test_no_keys(client):
    """Test that standard kvlayer APIs work correctly when not passed keys"""
    client.setup_namespace({'t1': 1})
    u = (uuid.uuid4(),)
    # this is a bug in several backend responses but we'll accept it for now
    uu = [u[0]]
    client.put('t1', (u, 'value'))
    assert list(client.get('t1', u)) == [(u, 'value')]
    assert list(client.scan('t1')) in [[(u, 'value')], [(uu, 'value')]]

    client.delete('t1')
    assert list(client.scan('t1')) in [[(u, 'value')], [(uu, 'value')]]

    client.get('t1')
    assert list(client.scan('t1')) in [[(u, 'value')], [(uu, 'value')]]

    client.put('t1')
    assert list(client.scan('t1')) in [[(u, 'value')], [(uu, 'value')]]


def test_new_key_spec(client):
    """Test that new table key specs work in setup_namespace."""
    client.setup_namespace({
        'kt2': (str, int),
        'kt3': (uuid.UUID, (int, long)),
        'kt4': (str, str, str),
    })

    good_kt2_key = ('aoeu', 1337)
    client.put('kt2', (good_kt2_key, 'blah'))

    with pytest.raises(BadKey):
        client.put('kt2', ((147, 'aeou'), 'nope'))
    with pytest.raises(BadKey):
        client.put('kt2', ((uuid.uuid4(), 'aeou'), 'nope'))
    with pytest.raises(BadKey):
        client.put('kt2', (('foo', 'aeou'), 'nope'))
    with pytest.raises(BadKey):
        client.put('kt2', (('foo', 123987419234701239847120), 'nope'))

    kt3u = uuid.uuid4()
    good_kt3_kvs = [
        ((kt3u, 42), 'v1'),
        ((kt3u, 0x7fFFFFffffFFFF), 'v2'),
    ]
    client.put('kt3', *good_kt3_kvs)

    count = 0
    for k, v in client.scan('kt3', ((kt3u,), (kt3u,))):
        assert (k, v) in good_kt3_kvs
        count += 1
    assert count == len(good_kt3_kvs)

    good_kt4_kvs = [
        (('fooa', 'aoeu', 'snth'), 'v1'),
        (('foob', 'aoeu', 'snth'), 'v1'),
        (('fooc', 'aoeu', 'snth'), 'v1'),
    ]
    client.put('kt4', *good_kt4_kvs)
    count = 0
    for k, v in client.scan('kt4', (('foo',), ('fooz',))):
        assert (k, v) in good_kt4_kvs
        count += 1
    assert count == len(good_kt4_kvs)


def test_merge_join(client):
    '''Test that we can read two tables at the same time.'''
    client.setup_namespace({
        't1': (str,),
        't2': (str,),
    })
    client.put('t1',
               (('a',), 'one'),
               (('b',), 'two'),
               (('d',), 'four'))
    client.put('t2',
               (('a',), 'one'),
               (('c',), 'three'))
    # Find keys in t1 not in t2
    iter1 = client.scan('t1')
    iter2 = client.scan('t2')
    missing = []
    v1 = next(iter1, None)
    v2 = next(iter2, None)
    while v1:
        if not v2:
            missing.append(v1)
            v1 = next(iter1, None)
        elif v1[0] < v2[0]:
            missing.append(v1)
            v1 = next(iter1, None)
        elif v1[0] == v2[0]:
            v1 = next(iter1, None)
            v2 = next(iter2, None)
        else:
            v2 = next(iter2, None)
    assert missing == [(('b',), 'two'), (('d',), 'four')]


def test_partial_scan(client):
    '''Test that reading part of a table, then starting a new scan,
    doesn't break.'''
    client.setup_namespace({'t1': (str,)})
    client.put('t1',
               (('a',), 'one'),
               (('b',), 'two'),
               (('c',), 'three'),
               (('d',), 'four'))

    s1 = client.scan('t1')
    assert next(s1, None) == (('a',), 'one')
    assert next(s1, None) == (('b',), 'two')

    s2 = client.scan('t1')
    assert next(s2, None) == (('a',), 'one')
    assert next(s2, None) == (('b',), 'two')
    assert next(s2, None) == (('c',), 'three')
    assert next(s2, None) == (('d',), 'four')
    assert next(s2, None) is None


def test_partial_scan_keys(client):
    '''Test that reading part of a table, then starting a new scan,
    doesn't break.'''
    client.setup_namespace({'t1': (str,)})
    client.put('t1',
               (('a',), 'one'),
               (('b',), 'two'),
               (('c',), 'three'),
               (('d',), 'four'))

    s1 = client.scan('t1')
    assert next(s1, None) == (('a',), 'one')
    assert next(s1, None) == (('b',), 'two')

    s2 = client.scan_keys('t1')
    assert next(s2, None) == ('a',)
    assert next(s2, None) == ('b',)
    assert next(s2, None) == ('c',)
    assert next(s2, None) == ('d',)
    assert next(s2, None) is None


def test_scan_name_oddity(client):
    client.setup_namespace({'index': (str, str, str)})
    row = (('NAME', 'alistair', 'vid'), '1')
    client.put('index', row)
    s = client.scan('index', (('NAME', 'a'), ('NAME', 'a\xff')))
    assert list(s) == [row]
    s = client.scan('index', (('NAME', 'al'), ('NAME', 'al\xff')))
    assert list(s) == [row]
    s = client.scan('index', (('NAME', 'ali'), ('NAME', 'ali\xff')))
    assert list(s) == [row]
    s = client.scan('index', (('NAME', 'alis'), ('NAME', 'alis\xff')))
    assert list(s) == [row]
    s = client.scan('index', (('NAME', 'alist'), ('NAME', 'alist\xff')))
    assert list(s) == [row]


def test_get_returns_keys(client):
    client.setup_namespace({'t': (str,)})
    key = ('fubar',)
    client.put('t', (key, '1'))
    results = list(client.get('t', key))
    assert len(results) == 1
    k, v = results[0]
    assert k == key
    assert v == '1'


def test_key_escaping(client):
    client.setup_namespace({'t': (str, str, str)})
    key = ('a\0b', 'a%b', 'a%00b')
    client.put('t', (key, '1'))
    val = list(client.get('t', key))
    assert len(val) == 1
    k, v = val[0]
    assert k == key
    assert v == '1'


# Value types

def test_int_value(client):
    client.setup_namespace({'t': (str,)}, {'t': int})
    client.put('t', (('k',), 1))
    assert list(client.get('t', ('k',))) == [(('k',), 1)]
    assert list(client.scan('t')) == [(('k',), 1)]
    assert isinstance(list(client.get('t', ('k',)))[0][1], int)


def test_float_value(client):
    client.setup_namespace({'t': (str,)}, {'t': float})
    client.put('t', (('k',), 1.0))
    assert list(client.get('t', ('k',))) == [(('k',), 1.0)]
    assert list(client.scan('t')) == [(('k',), 1.0)]
    assert isinstance(list(client.get('t', ('k',)))[0][1], float)


# Counters

def test_counter_minimal(client):
    client.setup_namespace({'t': (str,)}, {'t': kvlayer.COUNTER})
    client.increment('t', (('k',), 1))
    assert list(client.get('t', ('k',))) == [(('k',), 1)]
    assert list(client.scan('t')) == [(('k',), 1)]

    client.increment('t', (('k',), 1))
    assert list(client.get('t', ('k',))) == [(('k',), 2)]
    assert list(client.scan('t')) == [(('k',), 2)]

    client.increment('t', (('k',), 1))
    assert list(client.get('t', ('k',))) == [(('k',), 3)]
    assert list(client.scan('t')) == [(('k',), 3)]

    client.increment('t', (('k',), -3))
    assert list(client.get('t', ('k',))) == [(('k',), 0)]
    assert list(client.scan('t')) == [(('k',), 0)]

    assert list(client.get('t', ('x',))) == [(('x',), None)]


# (Because Accumulo has an interesting implementation of this)
def test_counter_put(client):
    client.setup_namespace({'t': (str,)}, {'t': kvlayer.COUNTER})
    client.put('t', (('k',), 1))
    assert list(client.get('t', ('k',))) == [(('k',), 1)]

    client.increment('t', (('k',), 1))
    assert list(client.get('t', ('k',))) == [(('k',), 2)]

    client.put('t', (('k',), 17))
    assert list(client.get('t', ('k',))) == [(('k',), 17)]

    client.increment('t', (('k',), -20))
    assert list(client.get('t', ('k',))) == [(('k',), -3)]



'''Tuple-to-string encoders.

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

This module provides converters between :mod:`kvlayer` key tuples and
flat byte strings.  These are needed by some backends that only store
single string keys.

.. automodule:: kvlayer.encoders.ascii_percent

'''
from __future__ import absolute_import

from kvlayer.encoders.ascii_percent import AsciiPercentEncoder
from kvlayer.encoders.packed import PackedEncoder

def get_encoder(name=None):
    '''Get an encoder instance by name.

    If `name` is :const:`None`, use a default encoder.

    :param str name: name of the encoder
    :return: instance of the encoder
    :raise KeyError: if `name` is not a valid encoder
    
    '''
    if name is None:
        name = 'ascii_percent'
    encoders = {encoder.config_name: encoder
                for encoder in [AsciiPercentEncoder, PackedEncoder]}
    return encoders[name]()



'''performance tests around batch size

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import division, absolute_import
import argparse
import cProfile
import logging
import sys
import time

import pytest

import kvlayer
from kvlayer.tests.test_interface import client, backend  # fixture
import yakonfig

logger = logging.getLogger(__name__)


def scan_batch_size(client):
    client.setup_namespace(dict(t1=(int,)))

    one_mb = r' ' * 2**20
    total_inserts = 200
    tasks = [((x,), one_mb) for x in xrange(total_inserts)]

    for batch_size in range(1, 100):        
        start_time = time.time()
        for start in range(0, total_inserts, batch_size):
            client.put('t1', *tasks[start:start+batch_size])
        elapsed = time.time() - start_time
        rate = total_inserts / elapsed
        print('%d MB written in %.1f seconds --> '
              '%.1f MB/sec with batch_size=%d for storage_type=%s' % (
                  total_inserts, elapsed, rate, batch_size,
                  client._config['storage_type']))

    #assert len(ret_vals) == total_inserts

    '''
        start_time = time.time()
        count = 0
        for (found, u) in run_many(many_gets, ret_vals, 
                                   class_config=client._config,
                                   num_workers=num_workers,
                                   timeout=total_inserts * 5,
                                   profile=profile):
            if not found:
                raise Exception('failed to find %r' % u)
            count += 1
        elapsed = time.time() - start_time
        assert len(ret_vals) == total_inserts
        rate = total_inserts / elapsed
        logger.info('%d MB read in %.1f seconds --> '
                    '%.1f MB/sec across %d workers for storage_type=%s',
                    total_inserts, elapsed, rate, num_workers,
                    client._config['storage_type'])

    '''

def main():
    import argparse
    parser = argparse.ArgumentParser(description='')
    parser.add_argument('--storage_type', default='redis')
    parser.add_argument('--storage_address', nargs='?', dest='storage_addresses')
    modules = [yakonfig, kvlayer]
    args = yakonfig.parse_args(parser, modules)
    config = yakonfig.get_global_config()

    if not args.storage_addresses:
        args.storage_addresses = ['redis.diffeo.com:6379']

    config['kvlayer'].update({
        'storage_type': args.storage_type,
        'storage_addresses': args.storage_addresses,
        })
    client = kvlayer.client()
    scan_batch_size(client)

if __name__ == '__main__':
    main()



'''
Implementation of AbstractStorage using Cassandra

This software is released under an MIT/X11 open source license.

Copyright 2012-2015 Diffeo, Inc.

..deprecated:: 0.3.0
  Code changes to support non-UUID keys in all other backends broke
  Cassandra, and as of kvlayer 0.3.1 this module does not work.

'''

import uuid
import time
import random
import logging
import traceback
from collections import defaultdict
from kvlayer._utils import join_uuids, split_uuids
from kvlayer._abstract_storage import AbstractStorage
from kvlayer._utils import _requires_connection
from thrift.transport.TTransport import TTransportException

logger = logging.getLogger('kvlayer.CStorage')

## get the Cassandra client library
import pycassa
from pycassa import NotFoundException
from pycassa.pool import ConnectionPool
from pycassa.system_manager import SystemManager, SIMPLE_STRATEGY, \
    ASCII_TYPE, BYTES_TYPE
from pycassa.types import AsciiType


class LittleLogger(object):
    def __init__(self, name):
        self.name = name

    def __call__(self, dic):
        #logger.debug('event: %s %r' % (self.name, dic))
        return

class _PycassaListener(object):
    def __init__(self, storage):
        self.storage = storage
        for method_name in ['connection_checked_out', 'connection_created', 'connection_disposed', 'connection_recycled', 'pool_at_max',
                            'pool_disposed', 'server_list_obtained', 'connection_checked_in']:
            setattr(self, method_name, LittleLogger(method_name))

    def connection_failed(self, dic):
        logger.critical('connection_failed: %s:' % str(dic))
        #self.storage.recreate_pool()


class CStorage(AbstractStorage):
    """
    Cassandra storage implements a set of table-like structures using
    ColumnFamilies in a namespace:

    http://www.slideshare.net/edanuff/indexing-in-cassandra

    """
    def __init__(self, *args, **kwargs):
        super(CStorage, self).__init__(*args, **kwargs)
        self.storage_addresses = self._config['storage_addresses']
        self.max_consistency_delay = self._config.get('max_consistency_delay', 120)
        ## avoid switching servers when creating/deleting namespaces
        self._chosen_server = random.choice(self.storage_addresses)
        logger.info('CStorage(_chosen_server=%r', self._chosen_server)
        self.pool_size = self._config.get('connection_pool_size', 2)
        self._connected = False
        self.thrift_framed_transport_size_in_mb = self._config.get('thrift_framed_transport_size_in_mb', 15)
        self.pool = None
        self.tables = {}
        self._app_namespace = self._app_name + '_' + self._namespace

    def setup_namespace(self, table_names):
        if self.pool:
            self.pool.dispose()
            del self.pool
            self.pool = None
        start_connect_time = time.time()

        super(CStorage, self).setup_namespace(table_names)

        sm = SystemManager(self._chosen_server)
        try:
            sm.create_keyspace(
                self._app_namespace, SIMPLE_STRATEGY,
                {
                    'replication_factor': str(self._config.get('replication_factor', '1'))
                },
                )
        except pycassa.InvalidRequestException, exc:
            if exc.why.startswith('Keyspace names must be case-insensitively unique'):
                pass
            else:
                raise exc

        self._create_tables(table_names, sm=sm)
        self.wait_for_consistency(sm=sm)

        ## now that we are consistent, we can create a pool
        self.pool = ConnectionPool(self._app_namespace, self.storage_addresses,
                          max_retries=1000, pool_timeout=10, pool_size=2, timeout=120)
        self.pool.fill()
        self.pool.add_listener(_PycassaListener(self))

        for family in self._table_names:
            if family not in self.tables:
                self.tables[family] = self._get_cf(family)

        elapsed = time.time() - start_connect_time
        logger.info('took %.3f seconds to setup_namespace() ConnectionPool(%d)' % (
                elapsed, self.pool_size))

        ## indicated connection is established
        self._connected = True

    def _get_cf(self, cf_name):
        return pycassa.ColumnFamily(
                self.pool, cf_name,
                read_consistency_level=pycassa.ConsistencyLevel.ALL,
                write_consistency_level=pycassa.ConsistencyLevel.ALL,
                )

    def _create_tables(self, table_names, sm=None):
        if sm is None:
            sm = SystemManager(self._chosen_server)
        for family, num_uuids in table_names.items():
            #CompositeType(*[UUIDType() for i in range(num_uuids)])
            comparator = AsciiType()
            try:
                sm.create_column_family(
                    self._app_namespace, family, super=False,
                    key_validation_class = ASCII_TYPE,
                    default_validation_class = BYTES_TYPE,
                    comparator_type=comparator,
                    )
            except pycassa.InvalidRequestException, exc:
                if exc.why.startswith('Cannot add already existing column family'):
                    pass
                else:
                    raise exc

    def wait_for_consistency(self, sm=None):
        if sm is None:
            sm = SystemManager(self._chosen_server)
        start_consistency_delay = time.time()
        consistency_delay = 0
        while len(sm.describe_schema_versions()) > 1 and \
                consistency_delay < self.max_consistency_delay:
            consistency_delay = time.time() - start_consistency_delay
            if consistency_delay > 20:
                logger.warn('waited %.1f seconds for cluster-wide consistency %r' % (
                        consistency_delay, sm.describe_schema_versions()))
            time.sleep(0.2)
        logger.info('number of schemas in cluster: %d' % len(sm.describe_schema_versions()))

    def delete_namespace(self):
        sm = SystemManager(self._chosen_server)
        try:
            sm.drop_keyspace(self._app_namespace)
        except pycassa.InvalidRequestException, exc:
            if exc.why.startswith('Cannot drop non existing keyspace'):
                pass
            else:
                raise exc
        except TTransportException, exc:
            logger.critical('trapping: %s' % traceback.format_exc(exc))

        sm.close()

    def _shard_number(self, joined_key):
        ## use first three hex characters, i.e. 4096 shards
        return int(joined_key[:3], 16)

    def _make_shard_name(self, table_name, joined_key):
        '''
        create a prefix-based sharding of table_name across the C* cluster
        '''
        return '%s-%04d' % (table_name, self._shard_number(joined_key))

    def _make_shard_names(self, table_name, start, finish):
        '''
        generate all row names (shards) needed between start and finish
        '''
        start_shard  = self._shard_number(start)
        finish_shard = self._shard_number(finish)
        row_names = []
        for shard_num in range(start_shard, finish_shard + 1):
            row_names.append('%s-%04d' % (table_name, shard_num))
        return row_names

    @_requires_connection
    def clear_table(self, table_name):
        self.tables[table_name].truncate()

    @_requires_connection
    def put(self, table_name, *keys_and_values, **kwargs):
        batch_size = kwargs.pop('batch_size', None)
        tot_bytes = 0
        cur_bytes = 0
        tot_rows = 0
        cur_rows = 0
        num_uuids = self._table_names[table_name]
        start = time.time()
        logger.debug('starting save')
        batch = self.tables[table_name].batch(queue_size=batch_size)
        for key, blob in keys_and_values:
            self.check_put_key_value(key, blob, table_name, num_uuids)
            if len(blob) + cur_bytes >= self.thrift_framed_transport_size_in_mb * 2**19:
                logger.critical('len(blob)=%d + cur_bytes=%d >= thrift_framed_transport_size_in_mb/2 = %d'
                                % (len(blob), cur_bytes, self.thrift_framed_transport_size_in_mb * 2**19))
                ## convert to MB and then cut in half
                if cur_rows > 0:
                    logger.critical('pre-emptively sending only what has been batched, and will send this item in next batch.')
                    batch.send()
                cur_bytes = 0
                cur_rows = 0
            cur_bytes += len(blob)
            tot_bytes += len(blob)
            cur_rows += 1
            tot_rows += 1
            if not isinstance(key, tuple):
                ## for consistency, always make things join_uuids,
                ## even if only one
                key = (key,)
            joined_key = join_uuids(*key)
            row_name = self._make_shard_name(table_name, joined_key)
            if len(blob) >= self.thrift_framed_transport_size_in_mb * 2**19:
                logger.critical('len(blob)=%d >= thrift_framed_transport_size_in_mb / 2 = %d, so there is a risk that the total payload will exceed the full thrift_framed_transport_size_in_mb, and the only solution to this is to change Cassandra server-side config to allow larger frames...'
                                % (len(blob), self.thrift_framed_transport_size_in_mb * 2**19))
            batch.insert(row_name, {joined_key: blob})
            #logger.critical('saving %s %r %r' % (table_name, key, blob))
            if tot_rows % 500 == 0:
                logger.debug('num rows=%d, num MB=%d, thrift_framed_transport_size_in_mb=%d' % (
                    tot_rows, float(tot_bytes) / 2**20, self.thrift_framed_transport_size_in_mb))

        batch.send()

        elapsed = time.time() - start
        row_rate = float(tot_rows)  / elapsed
        MB_rate  = float(tot_bytes) / elapsed / 2**20

        logger.info('%s.insert(%d rows, %d bytes in %.1f sec --> %.1f rows/sec %.3f MBps' % (
                table_name, tot_rows, tot_bytes, elapsed, row_rate, MB_rate))

    @_requires_connection
    def scan(self, table_name, *key_ranges, **kwargs):
        kwargs.pop('batch_size', 100)
        if not key_ranges:
            ## get all columns
            key_ranges = [['', '']]
        # TODO: s/num_uuids/key_spec/g
        num_uuids = self._table_names[table_name]
        for start, finish in key_ranges:
            specific_key_range = bool( start or finish )
            if specific_key_range and start == finish and len(start) == num_uuids:
                logger.warn('doing a scan on a single element, what?')
                #logger.info('specific_key_range: %r %r' % (start, finish))
                assert len(start) == num_uuids
                joined_key = join_uuids(*start)
                columns = [joined_key]
                row_names = [self._make_shard_name(table_name, joined_key)]
                start = None
                finish = None
            else:
                columns = None
                # TODO: require_uuid and num_uuids is obsolete. fix.
                start = make_start_key(start, uuid_mode=self._require_uuid, num_uuids=num_uuids)
                finish = make_end_key(finish, uuid_mode=self._require_uuid, num_uuids=num_uuids)
                row_names = self._make_shard_names(table_name, start, finish)
            total_count = 0
            hit_empty = False
            for row_name in row_names:
                try:
                    for key, val in self._get_from_one_row(table_name, row_name, columns, start, finish, num_uuids):
                        assert len(key) == num_uuids
                        yield key, val
                        if start:
                            assert start <= join_uuids(*key)
                        if finish:
                            assert finish >= join_uuids(*key)

                        total_count += 1
                        #logger.critical('total_count: %d' % total_count)
                except pycassa.NotFoundException:
                    hit_empty = True

            #logger.debug('specific_key_range: %r  hit_empty: %r  total_count: %r' %
            #                (specific_key_range, hit_empty, total_count))

    def _get_from_one_row(self, table_name, row_name, columns, start, finish, num_uuids):

        logger.debug('c* get: table_name=%r row_name=%r columns=%r start=%r finish=%r' % (
            table_name, row_name, columns, start, finish))

        if not columns:
            assert start is not None and finish is not None
            assert start <= finish

        num_yielded = 0
        while True:
            ## if we have
            prev_start = start
            logger.debug('cassandra get(%r...)' % row_name)
            ## if table_name == 'inbound':
            ##     import ipdb
            ##     ipdb.set_trace()
            for key, val in self.tables[table_name].get(
                    row_name,
                    columns=columns,
                    column_start=start,
                    column_finish=finish,
                    column_count=1,
                    ).iteritems():

                key = split_uuids(key)
                logger.critical('cassandra get(%r) yielding %r %d' % (table_name, key, len(val)))
                yield key, val
                num_yielded += 1
                logger.debug('c* get: table_name=%r row_name=%r columns=%r start=%r finish=%r' % (
                    table_name, row_name, columns, start, finish))

            ## prepare to page ahead to next batch
            if columns:
                break
            start = list(key)
            start[-1] = uuid.UUID(int=key[-1].int+1)
            assert len(start) == num_uuids
            start = join_uuids(*start)

            if start == prev_start or start > finish:
                break
            logger.debug('paging forward from %r to %r' % (prev_start, start))

        ## We need to raise a not found exception if the caller asked for
        ## a specific column and we didn't yield any results
        if not columns and num_yielded == 0:
            raise pycassa.NotFoundException
            #'c* get: table_name=%r row_name=%r columns=%r start=%r finish=%r' % (
            #    table_name, row_name, columns, start, finish))


    @_requires_connection
    def get(self, table_name, *keys, **kwargs):
        num_uuids = self._table_names[table_name]
        ## A mapping between shards and columns to retrieve from each shard
        shards = defaultdict(list)

        ## Track the keys that we find
        found_keys = set()

        ## Determine all the shards that we need to contact
        for key in keys:
            assert len(key) == num_uuids
            joined_key = join_uuids(*key)
            row_names = [self._make_shard_name(table_name, joined_key)]
            for row_name in row_names:
                shards[row_name].append(joined_key)

        ## For each shard we contact, get all the keys it may hold
        for row_name, columns in shards.iteritems():
            try:
                for key, val in self.tables[table_name].get(row_name, columns=columns,).iteritems():
                    key = split_uuids(key)
                    logger.critical('cassandra get(%r) yielding %r %d' % (table_name, key, len(val)))
                    ## key is a list [(uuid, uuid)]
                    assert not any(k in found_keys for k in key)
                    found_keys.add(tuple(key))
                    yield tuple(key), val
            except NotFoundException:
                pass

        ## Raise an exception if we don't retrieve all the keys that were requested
        missing_keys = set(keys) - found_keys
        for k in missing_keys:
            yield k, None


    @_requires_connection
    def delete(self, table_name, *keys, **kwargs):
        num_uuids = self._table_names[table_name]
        batch_size = kwargs.pop('batch_size', 1000)
        batch = self.tables[table_name].batch(queue_size=batch_size)
        count = 0
        for key in keys:
            assert len(key) == num_uuids
            joined_key = join_uuids(*key)
            row_name = self._make_shard_name(table_name, joined_key)
            columns = [joined_key]
            #logger.critical('C* delete: table_name=%r columns=%r' % (table_name, columns))
            batch.remove(row_name, columns=columns)
            count += 1
        batch.send()
        logger.info('deleted %d tree_ids from %r' % (count, table_name))

    def close(self):
        self._connected = True
        if hasattr(self, 'pool') and self.pool:
            self.pool.dispose()

    def __del__(self):
        self.close()



'''
Implementation of AbstractStorage using python shelve module

Your use of this software is governed by your license agreement.

Copyright 2012-2014 Diffeo, Inc.
'''

from __future__ import absolute_import
import os
import shutil
import shelve
import logging
import cPickle
from kvlayer._local_memory import AbstractLocalStorage

logger = logging.getLogger(__name__)


class FileStorage(AbstractLocalStorage):
    '''
    File storage for testing and development
    '''

    _datas = {}

    def __init__(self, *args, **kwargs):
        super(FileStorage, self).__init__(*args, **kwargs)

        filename = self._config['filename']
        if filename in self._datas:
            self._data = self._datas[filename]
        else:
            if self._config.get('copy_to_filename', False):
                copy_to_filename = self._config['copy_to_filename']
                shutil.copyfile(filename, copy_to_filename)
                filename = copy_to_filename

            if os.path.exists(filename) and os.path.getsize(filename) == 0:
                os.remove(filename)
            self._data = shelve.open(filename,
                                     protocol=cPickle.HIGHEST_PROTOCOL,
                                     writeback=True)
            self._datas[filename] = self._data

        self._table_names = {}

    def delete_namespace(self):
        self._data.clear()

    def put(self, table_name, *keys_and_values, **kwargs):
        super(FileStorage, self).put(table_name, *keys_and_values, **kwargs)
        self._data.sync()

    def delete(self, table_name, *keys):
        super(FileStorage, self).delete(table_name, *keys)
        self._data.sync()
